//////////////////////////////////////////////////////////////////////
//
//  IrCOMM2k - Virtual Infrared COM Port for Windows 2000/XP
//  Copyright (c) 2001, 2002 Jan Kiszka
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//////////////////////////////////////////////////////////////////////

// Include-Datei mit Defines zum Bridge-Interface, wird auch vom Service-
// Programm eingebunden

// History: 1.0.0.49  - erste freigegebene Version
//          1.0.1.55  - IOCTL_IRCOMM_CANCEL_GET_COMMAND gestrichen
//          1.2.0.62  - IOCTL_IRCOMM_UPDATE_CONFIG eingefhrt
//          1.2.1.71  - IRCOMM_CMD_DEV_CLOSE eingefhrt

//
// Defines fr Bridge-Device
//
#define BRIDGE_DEVICE_TYPE          41911
#define BRIDGE_DEVICE_NAME_SERVICE  "\\\\.\\COMMBRDG"
#define BRIDGE_DEVICE_NAME_DRIVER   L"\\DosDevices\\COMMBRDG"


//
// IOCTL-Codes fr Bridge-Device
//
//  IOCTL_IRCOMM_CHECK_VERSION:
//      InputBuffer      = Zeiger auf ULONG-Variable, die die Build-Nummer des
//                         Service-Programms enthlt
//      InputBufferSize  = 4
//      OutputBuffer     = Zeiger auf ULONG-Variable, in der die Build-Nummer
//                         des Treibers zurckgeliefert wird
//      OutputBufferSize = 4
//  Rckgabe: Falls der Treiber nicht mit der angegebenen Version des Service-
//      Programms zusammenarbeiten kann, liefert die IOCTL-Funktion einen
//      Fehler zurck.
//
//  IOCTL_IRCOMM_GET_COMMAND:
//      InputBuffer      = NULL
//      InputBufferSize  = 0
//      OutputBuffer     = Zeiger auf ULONG-Variable, in der das Kommando
//                         zurckgeliefert wird
//      OutputBufferSize = 4
//
//  IOCTL_IRCOMM_REPORT_STATUS:
//      InputBuffer      = Zeiger auf ULONG-Variable mit neuem Status
//      InputBufferSize  = 4
//      OutputBuffer     = NULL
//      OutputBufferSize = 0
//
//  IOCTL_IRCOMM_UPDATE_CONFIG:
//      InputBuffer      = NULL
//      InputBufferSize  = 0
//      OutputBuffer     = NULL
//      OutputBufferSize = 0
//
#define IOCTL_IRCOMM_CHECK_VERSION      CTL_CODE(BRIDGE_DEVICE_TYPE, 1, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_IRCOMM_GET_COMMAND        CTL_CODE(BRIDGE_DEVICE_TYPE, 2, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_IRCOMM_REPORT_STATUS      CTL_CODE(BRIDGE_DEVICE_TYPE, 3, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_IRCOMM_UPDATE_CONFIG      CTL_CODE(BRIDGE_DEVICE_TYPE, 4, METHOD_BUFFERED, FILE_ANY_ACCESS)


//
// Kommandos von IOCTL_IRCOMM_GET_COMMAND
//
#define IRCOMM_CMD_IR_CONNECT           1
#define IRCOMM_CMD_IR_DISCONNECT        2
#define IRCOMM_CMD_DEV_CLOSE            3


//
// Statusmeldungen fr IOCTL_IRCOMM_REPORT_STATUS
//
#define IRCOMM_STATUS_CONNECTED         1
#define IRCOMM_STATUS_DISCONNECTED      2
#define IRCOMM_STATUS_NO_DEVICES_FOUND  3
