#ifndef _LINUX_SPINLOCK_H
#define _LINUX_SPINLOCK_H

#include "stdafx.h"

typedef HANDLE spinlock_t;

extern HANDLE globalLock;

inline void spin_lock_init(spinlock_t* pSpinlock)
{
    *pSpinlock = CreateMutex(NULL, false, NULL);
}

#define spin_lock_irqsave(pLock, flags) do \
{ \
    flags = 0; \
    WaitForSingleObject(*pLock, INFINITE); \
} while (0)

#define spin_unlock_irqrestore(pLock, flags) do \
{ \
    ReleaseMutex(*pLock); \
} while (0)

#define SPIN_LOCK_UNLOCKED  CreateMutex(NULL, false, NULL)

#define save_flags(flags) do \
{ \
    flags = 0; \
    WaitForSingleObject(globalLock, INFINITE); \
} while (0)

#define cli()
#define restore_flags(x) do \
{ \
    ReleaseMutex(globalLock); \
} while (0)

#define test_bit(bit, x)    (*(x) & (1 << bit))

__inline void set_bit(int nr, __u32* addr)
{
    __asm
    {
        mov  eax,nr
        mov  ebx,addr
        bts  [ebx],eax
    }
}

__inline void clear_bit(int nr, __u32* addr)
{
    __asm
    {
        mov  eax,nr
        mov  ebx,addr
        btr  [ebx],eax
    }
}

__inline int test_and_clear_bit(int nr, __u32* addr)
{
    int result;


    __asm
    {
        mov  eax,nr
        mov  ebx,addr
        btr  [ebx],eax
        sbb  eax,eax
        mov  result,eax
    }
    return result;
}

__inline int test_and_set_bit(int nr, __u32* addr)
{
    int result;


    __asm
    {
        mov  eax,nr
        mov  ebx,addr
        bts  [ebx],eax
        sbb  eax,eax
        mov  result,eax
    }
    return result;
}

#endif