#ifndef _LINUX_TIMER_H
#define _LINUX_TIMER_H

#include <linux/types.h>

#define HZ          1000
#define jiffies     GetTickCount()

struct timer_list {
    unsigned long expires;
    unsigned long data;
    void (*function)(unsigned long);

    // Win32 specific
    struct timer_list* pNext;
    HANDLE  hTimerThread;
    HANDLE  hStartEvent;
    HANDLE  hTimeoutEvent;
};

extern struct timer_list* pTimers;

extern DWORD WINAPI TimerThread(LPVOID arg);

__inline void init_timer(struct timer_list* timer)
{
    timer->pNext = pTimers;
    pTimers      = timer;

    timer->hStartEvent   = CreateEvent(NULL, true, false, NULL);
    timer->hTimeoutEvent = CreateEvent(NULL, false, false, NULL);
    timer->hTimerThread  = CreateThread(NULL, 0, TimerThread, timer, 0, NULL);
}

__inline void add_timer(struct timer_list* timer)
{
    timer->expires += 100-jiffies;
    ResetEvent(timer->hTimeoutEvent);
    SetEvent(timer->hStartEvent);
}

__inline int del_timer(struct timer_list* timer)
{
    SetEvent(timer->hTimeoutEvent);
    return 0;
}

__inline int timer_pending (const struct timer_list* timer)
{
    return (WaitForSingleObject(timer->hTimeoutEvent, 0) == WAIT_TIMEOUT);
}

#endif