#include "stdafx.h"

#include "brdgdev.h"

#include <asm/byteorder.h>

#include <linux/config.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/skbuff.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/random.h>
#include <linux/if_ether.h>
#include <linux/sched.h>

#include <net/irda/crc.h>
#include <net/irda/irda.h>
#include <net/irda/irda_device.h>
#include <net/irda/irlap.h>
#include <net/irda/irlmp.h>
#include <net/irda/iriap.h>
#include <net/irda/irttp.h>
#include <net/irda/ircomm_tty.h>
#include <net/irda/ircomm_tty_attach.h>
#include <net/irda/wrapper.h>

typedef struct {
    HANDLE      events[7];
    HANDLE      bridgeDev;
    OVERLAPPED  ovBridgeIOCTL;
    DWORD       command;
} CONNECTION_CONTEXT, *PCONNECTION_CONTEXT;

HANDLE                  hComPort;
BOOL                    stop = false;
BOOL                    mediaBusy = false;
struct timer_list*      pTimers = NULL;
HANDLE                  globalLock;
HANDLE                  xmitLock;
extern char             sysctl_devname[]; 
struct qos_info         qos;
struct net_device       netdev;
struct irlap_cb*        pIrlap;
hashbin_t*              ircomm_tty = NULL;
struct tty_struct       tty;
__u32                   irda_debug = 2;
__u8                    xmitBuf[5000];
OVERLAPPED              writeOvr;
DCB                     dcb;
CONNECTION_CONTEXT      connContext;

__u16 const irda_crc16_table[256] =
{
	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

extern int irlap_driver_rcv(struct sk_buff*, struct net_device*,
                            struct packet_type*);
extern int ircomm_init(void);
extern int ircomm_tty_startup(struct ircomm_tty_cb *self);
extern void ircomm_tty_shutdown(struct ircomm_tty_cb *self);

inline void async_bump(struct net_device *dev, struct net_device_stats *stats,
                       __u8 *buf, int len)
{
    struct sk_buff* skb;
    int result;


    skb = dev_alloc_skb(len);
    if (skb == NULL)
        return;

    memcpy(skb_put(skb, len-2), buf, len-2);

    skb->dev = dev;
    skb->mac.raw  = skb->data;
    skb->protocol = htons(ETH_P_IRDA);

    printf("==> ");
    while (len-- > 0)
        printf("%02X ", *buf++);
    printf("\n");
    if ((result = irlap_driver_rcv(skb, dev, NULL)) != 0)
        printf("irlap_driver_rcv = %d\n", result);
}

static void state_outside_frame(iobuff_t *rx_buff, __u8 byte)
{
    switch (byte)
    {
        case BOF:
            rx_buff->state = BEGIN_FRAME;
            rx_buff->in_frame = TRUE;
            break;
        case XBOF:
            /* idev->xbofs++; */
            break;
        case EOF:
        default:
            //irda_device_set_media_busy(dev, TRUE);
            mediaBusy = true;
            break;
    }
}

static void state_begin_frame(iobuff_t *rx_buff, __u8 byte)
{
    /* Time to initialize receive buffer */
    rx_buff->data = rx_buff->head;
    rx_buff->len = 0;
    rx_buff->fcs = INIT_FCS;

    switch (byte)
    {
        case BOF:
            /* Continue */
            break;
        case CE:
            /* Stuffed byte */
            rx_buff->state = LINK_ESCAPE;
            break;
        case EOF:
            /* Abort frame */
            rx_buff->state = OUTSIDE_FRAME;
            printf("^abort frame\n");
            break;
        default:
            rx_buff->data[rx_buff->len++] = byte;
            rx_buff->fcs = irda_fcs(rx_buff->fcs, byte);
            rx_buff->state = INSIDE_FRAME;
            break;
    }
}

static void state_link_escape(iobuff_t *rx_buff, __u8 byte)
{
    switch (byte)
    {
        case BOF: /* New frame? */
            printf("^Discarding incomplete frame\n");
            rx_buff->state = BEGIN_FRAME;
            //irda_device_set_media_busy(dev, TRUE);
            mediaBusy = true;
            break;
        case CE:
            printf("^state not defined\n");
            break;
        case EOF: /* Abort frame */
            rx_buff->state = OUTSIDE_FRAME;
            break;
        default:
            /* 
             *  Stuffed char, complement bit 5 of byte 
             *  following CE, IrLAP p.114 
             */
            byte ^= IRDA_TRANS;
            if (rx_buff->len < rx_buff->truesize)
            {
	        rx_buff->data[rx_buff->len++] = byte;
	        rx_buff->fcs = irda_fcs(rx_buff->fcs, byte);
	        rx_buff->state = INSIDE_FRAME;
            }
            else
            {
	        printf("^rx buffer overflow\n");
	        rx_buff->state = OUTSIDE_FRAME;
            }
            break;
    }
}

static void state_inside_frame(iobuff_t *rx_buff, __u8 byte)
{
    int ret = 0; 

    switch (byte)
    {
        case BOF: /* New frame? */
            printf("^Discarding incomplete frame\n");
            rx_buff->state = BEGIN_FRAME;
            //irda_device_set_media_busy(dev, TRUE);
            mediaBusy = true;
            break;
        case CE: /* Stuffed char */
            rx_buff->state = LINK_ESCAPE;
            break;
        case EOF: /* End of frame */
            rx_buff->state = OUTSIDE_FRAME;
            rx_buff->in_frame = FALSE;

            /* Test FCS and signal success if the frame is good */
            if (rx_buff->fcs == GOOD_FCS)
            {
                /* Deliver frame */
                async_bump(&netdev, NULL, rx_buff->data, rx_buff->len);
                ret = TRUE;
                break;
            }
            else 
            {
                /* Wrong CRC, discard frame!  */
                //irda_device_set_media_busy(dev, TRUE); 
                mediaBusy = true;

                printf("^crc error\n");
            }			
            break;
        default: /* Must be the next byte of the frame */
            if (rx_buff->len < rx_buff->truesize)
            {
                rx_buff->data[rx_buff->len++] = byte;
                rx_buff->fcs = irda_fcs(rx_buff->fcs, byte);
            }
            else
            {
                printf("^Rx buffer overflow, aborting\n");
                rx_buff->state = OUTSIDE_FRAME;
            }
            break;
    }
}
 
static void (*state[])(iobuff_t *rx_buff, __u8 byte) = 
{ 
    state_outside_frame,
    state_begin_frame,
    state_link_escape,
    state_inside_frame
};

inline void async_unwrap_char(struct net_device *dev, struct net_device_stats *stats,
                              iobuff_t *rx_buff, __u8 byte)
{
    (*state[rx_buff->state])(rx_buff, byte);
}

static inline int stuff_byte(__u8 byte, __u8 *buf)
{
    switch (byte)
    {
        case BOF: /* FALLTHROUGH */
        case EOF: /* FALLTHROUGH */
        case CE:
            /* Insert transparently coded */
            buf[0] = CE;               /* Send link escape */
            buf[1] = byte^IRDA_TRANS;    /* Complement bit 5 */
            return 2;
            /* break; */
        default:
             /* Non-special value, no transparency required */
            buf[0] = byte;
            return 1;
            /* break; */
    }
}

int async_wrap_skb(struct sk_buff *skb, __u8 *tx_buff, int buffsize)
{
    struct irda_skb_cb *cb = (struct irda_skb_cb *) skb->cb;
    int xbofs;
    unsigned int i;
    int n;
    union {
        __u16 value;
        __u8 bytes[2];
    } fcs;

    /* Initialize variables */
    fcs.value = INIT_FCS;
    n = 0;

    /*
     *  Send  XBOF's for required min. turn time and for the negotiated
     *  additional XBOFS
     */
	
    if (cb->magic != LAP_MAGIC)
    {
        /* 
         * This will happen for all frames sent from user-space.
         * Nothing to worry about, but we set the default number of 
         * BOF's
         */
	printf("^wrong magic in skb!\n");
	xbofs = 10;
    }
    else
        xbofs = cb->xbofs + cb->xbofs_delay;

    /* Check that we never use more than 115 + 48 xbofs */
    if (xbofs > 163) {
        printf("^too many xbofs (%d)\n", xbofs);
        xbofs = 163;
    }

    memset(tx_buff+n, XBOF, xbofs);
    n += xbofs;

    /* Start of packet character BOF */
    tx_buff[n++] = BOF;

    /* Insert frame and calc CRC */
    for (i=0; i < skb->len; i++)
    {
        /*
         *  Check for the possibility of tx buffer overflow. We use
         *  bufsize-5 since the maximum number of bytes that can be 
         *  transmitted after this point is 5.
         */
        ASSERT(n < (buffsize-5), return n;);

        n += stuff_byte(skb->data[i], tx_buff+n);
        fcs.value = irda_fcs(fcs.value, skb->data[i]);
    }
	
    /* Insert CRC in little endian format (LSB first) */
    fcs.value = ~fcs.value;
    n += stuff_byte(fcs.bytes[0], tx_buff+n);
    n += stuff_byte(fcs.bytes[1], tx_buff+n);
    tx_buff[n++] = EOF;

    return n;
}

int dev_queue_xmit(struct sk_buff* skb)
{
    int len;
    int i;
    __u32 newSpeed;
    static __u32 baudRate[] = {9600, 19200, 57600, 115200, 38400};
    static __u32 devSpeed = 0;


    if (skb->len > 0)
    {
        {
            __u8* buf;
            len = skb->len;
            buf = skb->data;
            printf("<== ");
            while (len-- > 0)
                printf("%02X ", *buf++);
            printf("\n");
        }

        WaitForSingleObject(xmitLock, INFINITE);

        len = async_wrap_skb(skb, xmitBuf, sizeof(xmitBuf));
        WriteFile(hComPort, xmitBuf, len, NULL, &writeOvr);
        WaitForSingleObject(writeOvr.hEvent, INFINITE);

        ReleaseMutex(xmitLock);
    }

    newSpeed = ((struct irlap_cb*)skb->dev->atalk_ptr)->speed;
    if (newSpeed != devSpeed)
    {
        FlushFileBuffers(hComPort);
        Sleep(50);

        dcb.BaudRate = newSpeed;
        SetCommState(hComPort, &dcb);

        EscapeCommFunction(hComPort, CLRDTR);
        Sleep(1);
        EscapeCommFunction(hComPort, SETDTR);

        for (i = 0; (newSpeed != baudRate[i]) && (i < 5); i++)
        {
            Sleep(1);
            EscapeCommFunction(hComPort, CLRRTS);
            Sleep(1);
            EscapeCommFunction(hComPort, SETRTS);
        }

        devSpeed = newSpeed;
    }

    kfree_skb(skb);

    return 0;
}

DWORD WINAPI TimerThread(LPVOID arg)
{
    struct timer_list*  pTimerData  = (struct timer_list*) arg;
    struct timer_list*  pNext;


    pNext = pTimerData->pNext;
    while (pNext == pTimerData->pNext)
    {
        WaitForSingleObject(pTimerData->hStartEvent, INFINITE);
        if (pNext != pTimerData->pNext)
            return 0;
        ResetEvent(pTimerData->hStartEvent);

        if (WaitForSingleObject(pTimerData->hTimeoutEvent,
                                pTimerData->expires) == WAIT_TIMEOUT)
        {
            if (pNext != pTimerData->pNext)
                return 0;
            pTimerData->function(pTimerData->data);
        }
    }
    return 0;
}

void CleanupTimers()
{
    while (pTimers != NULL)
    {
        TerminateThread(pTimers->hTimerThread, 0);
        CloseHandle(pTimers->hTimerThread);
        CloseHandle(pTimers->hStartEvent);
        CloseHandle(pTimers->hTimeoutEvent);
        pTimers->hTimeoutEvent = INVALID_HANDLE_VALUE;

        pTimers = pTimers->pNext;
    }
}

void OpenPort()
{
    struct ircomm_tty_cb* self;
    int ret;
    struct timer_list tmp;


    self = hashbin_find(ircomm_tty, /*line*/0, NULL);
    if (self->magic == IRCOMM_TTY_MAGIC)
        return;

    memcpy(&tmp, &self->watchdog_timer, sizeof(struct timer_list));
    memset(((__u8*)self)+sizeof(irda_queue_t), 0, sizeof(struct ircomm_tty_cb)-sizeof(irda_queue_t));
    memcpy(&self->watchdog_timer, &tmp, sizeof(struct timer_list));

    self->magic = IRCOMM_TTY_MAGIC;
    self->flow = FLOW_STOP;

    self->line = /*line*/0;
//    self->tqueue.routine = ircomm_tty_do_softint;
//    self->tqueue.data = self;
    self->max_header_size = 5;
    self->max_data_size = 64-self->max_header_size;
    self->close_delay = 5*HZ/10;
    self->closing_wait = 30*HZ;

//    init_waitqueue_head(&self->open_wait);
//    init_waitqueue_head(&self->close_wait);

    self->open_count++;

    self->tty = &tty;

    if (/*line*/0 < 0x10)
    {
        self->service_type = IRCOMM_3_WIRE | IRCOMM_9_WIRE;
        self->settings.service_type = IRCOMM_9_WIRE; /* 9 wire as default */
        self->settings.dce = IRCOMM_CTS | IRCOMM_CD; /* Default line settings */
    }
    else
    {
        self->service_type = IRCOMM_3_WIRE_RAW;
        self->settings.service_type = IRCOMM_3_WIRE_RAW; /* Default */
    }

    ret = ircomm_tty_startup(self);

}

void ClosePort()
{
    struct ircomm_tty_cb* self;


    self = hashbin_find(ircomm_tty, /*line*/0, NULL);
    if (self->magic != IRCOMM_TTY_MAGIC)
        return;

    self->open_count--;
    self->flags |= ASYNC_CLOSING;

    ircomm_tty_shutdown(self);

    self->magic = ~IRCOMM_TTY_MAGIC;
}

void IrCOMMWrite(char* buf, unsigned int len)
{
    struct ircomm_tty_cb* self;
    struct sk_buff *skb;
    unsigned int blockLen;


    self = hashbin_find(ircomm_tty, /*line*/0, NULL);
    if (self->magic != IRCOMM_TTY_MAGIC)
        return;

    while (len > 0)
    {
        if (len > self->max_data_size)
            blockLen = self->max_data_size;
        else
            blockLen = len;

        skb = dev_alloc_skb(self->max_data_size + self->max_header_size);
        skb_reserve(skb, self->max_header_size);

        memcpy(skb_put(skb, blockLen), buf, blockLen);
        ircomm_tty_do_event(self, IRCOMM_TTY_DATA_REQUEST, skb, NULL);

        len -= blockLen;
        buf += blockLen;
    }
}

DWORD WINAPI ConnectionThread(LPVOID lpParameter)
{
    #define BRIDGE_BUFFER_SIZE              65536
    #define EVENT_IOCTL_BRIDGE              0
    #define EVENT_READ_BRIDGE               1
    #define IRCOMM_ERROR_MALLOC             1000000
    #define IRCOMM_ERROR_BRIDGE_REMOVED     1000003
//#define EVENT_WRITE_BRIDGE              6
    PCONNECTION_CONTEXT pConnContext  = (PCONNECTION_CONTEXT)lpParameter;
//    SOCKET              servSock      = INVALID_SOCKET;
//    SOCKET              connSock      = INVALID_SOCKET;
    void*               bufferIR2Brdg = NULL;
    void*               bufferBrdg2IR = NULL;
    DWORD               flags         = 0;
    BOOLEAN             IRConnected   = FALSE;
    int                 returnCode;
//    DWORD               transferIR2Brdg;
    DWORD               transferBrdg2IR;
    DWORD               IOCTLSize;
//    OVERLAPPED          ovReadIR;
//    OVERLAPPED          ovWriteIR;
    OVERLAPPED          ovReadBridge;
//    OVERLAPPED          ovWriteBridge;
    DWORD               status;
//    UINT                svcControl;
    int                 result;
//    int                 i;


    connContext.events[EVENT_IOCTL_BRIDGE] = CreateEvent(NULL, TRUE, FALSE, NULL);
    connContext.events[EVENT_READ_BRIDGE] = CreateEvent(NULL, TRUE, FALSE, NULL);
    connContext.ovBridgeIOCTL.hEvent = connContext.events[EVENT_IOCTL_BRIDGE];
    connContext.bridgeDev = CreateFile(BRIDGE_DEVICE_NAME_SERVICE, GENERIC_READ+GENERIC_WRITE, 0,
                                       NULL, OPEN_EXISTING,
                                       FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, NULL);
    if (!DeviceIoControl(connContext.bridgeDev, IOCTL_IRCOMM_GET_COMMAND, NULL, 0,
                         &connContext.command, sizeof(connContext.command), NULL,
                         &connContext.ovBridgeIOCTL))
    {
        result = GetLastError();
        if (result != ERROR_IO_PENDING)
        {
            result = GetLastError();
            goto ConnectionCleanup;
        }
    }
    else
        SetEvent(connContext.events[EVENT_IOCTL_BRIDGE]);

    // Events EVENT_READ_IR bis EVENT_WRITE_BRIDGE zurcksetzen
//    for (i = 2; i < 7; i++)
//        ResetEvent(pConnContext->events[i]);

    // Overlapped-Strukturen initialisieren
//    memset(&ovReadIR, 0, sizeof(ovReadIR));
//    ovReadIR.hEvent      = pConnContext->events[EVENT_READ_IR];
//    memset(&ovWriteIR, 0, sizeof(ovWriteIR));
//    ovWriteIR.hEvent     = pConnContext->events[EVENT_WRITE_IR];
    memset(&ovReadBridge, 0, sizeof(ovReadBridge));
    ovReadBridge.hEvent  = pConnContext->events[EVENT_READ_BRIDGE];
//    memset(&ovWriteBridge, 0, sizeof(ovWriteBridge));
//    ovWriteBridge.hEvent = pConnContext->events[EVENT_WRITE_BRIDGE];

    // Puffer reservieren (2 x 4 KByte, Device-Liste)
    bufferIR2Brdg = malloc(2*BRIDGE_BUFFER_SIZE);
    if (bufferIR2Brdg == NULL)
    {
        returnCode = IRCOMM_ERROR_MALLOC;
        goto ConnectionCleanup;
    }

    bufferBrdg2IR = (BYTE*)bufferIR2Brdg + BRIDGE_BUFFER_SIZE;

/*    pDevList = (PDEVICELIST)malloc(DEV_LIST_BUFFER_SIZE);
    if (pDevList == NULL)
    {
        returnCode = IRCOMM_ERROR_MALLOC;
        goto ConnectionCleanup;
    }*/

    while (TRUE)
    {
        switch (WaitForMultipleObjects(2, pConnContext->events, FALSE, INFINITE))
        {
            // Kommando vom Treiber
            case WAIT_OBJECT_0 + EVENT_IOCTL_BRIDGE:
                ResetEvent(pConnContext->events[EVENT_IOCTL_BRIDGE]);

                if (GetOverlappedResult(pConnContext->bridgeDev, &pConnContext->ovBridgeIOCTL,
                                        &IOCTLSize, FALSE) &&
                    (IOCTLSize == sizeof(pConnContext->command)))
                {
                    if (!DeviceIoControl(pConnContext->bridgeDev,
                                         IOCTL_IRCOMM_GET_COMMAND, NULL, 0,
                                         &pConnContext->command,
                                         sizeof(pConnContext->command), &IOCTLSize,
                                         &pConnContext->ovBridgeIOCTL))
                    {
                        result = GetLastError();
                        if (result != ERROR_IO_PENDING)
                        {
                            returnCode = result;
                            goto ConnectionCleanup;
                        }
                    }
                    else
                        SetEvent(pConnContext->events[EVENT_IOCTL_BRIDGE]);

                    switch (pConnContext->command)
                    {
                        case IRCOMM_CMD_IR_CONNECT:
                            if (!IRConnected)
                            {
                                IRConnected = TRUE;

                                OpenPort();

                                ovReadBridge.Offset     = 0;
                                ovReadBridge.OffsetHigh = 0;
                                if (!ReadFile(pConnContext->bridgeDev, bufferBrdg2IR,
                                              BRIDGE_BUFFER_SIZE, &transferBrdg2IR,
                                              &ovReadBridge))
                                {
                                    result = GetLastError();
                                    if (result != ERROR_IO_PENDING)
                                    {
                                        returnCode = result;
                                        goto ConnectionCleanup;
                                    }
                                }
                                else
                                    SetEvent(pConnContext->events[EVENT_READ_BRIDGE]);

                                // Benachrichtigung an Bridge-Device senden
                                status = IRCOMM_STATUS_CONNECTED;
                                DeviceIoControl(pConnContext->bridgeDev,
                                                IOCTL_IRCOMM_REPORT_STATUS, &status,
                                                sizeof(status), NULL, 0,
                                                &IOCTLSize, NULL);
                            }

                            break;

                        case IRCOMM_CMD_IR_DISCONNECT:
                            if (IRConnected)
                            {
                                ClosePort();

                                IRConnected = FALSE;

                                // Benachrichtigung an Bridge-Device senden
                                status = IRCOMM_STATUS_DISCONNECTED;
                                DeviceIoControl(pConnContext->bridgeDev,
                                                IOCTL_IRCOMM_REPORT_STATUS, &status,
                                                sizeof(status), NULL, 0, &IOCTLSize, NULL);
                            }

                            returnCode = 0;
//                            goto ConnectionCleanup;
                            break;

                        case IRCOMM_CMD_DEV_CLOSE:
                            returnCode = IRCOMM_ERROR_BRIDGE_REMOVED;
                            goto ConnectionCleanup;
                    }
                }
                else
                {
                    returnCode = GetLastError();
                    goto ConnectionCleanup;
                }

                break;

/*            // Lesen von IR
            case WAIT_OBJECT_0 + EVENT_READ_IR:
                ResetEvent(pConnContext->events[EVENT_READ_IR]);

                if (GetOverlappedResult((HANDLE)connSock, &ovReadIR, &transferIR2Brdg, FALSE))
                {
                    ovWriteBridge.Offset     = 0;
                    ovWriteBridge.OffsetHigh = 0;
                    if (!WriteFile(pConnContext->bridgeDev, bufferIR2Brdg, transferIR2Brdg,
                                   &transferIR2Brdg, &ovWriteBridge))
                    {
                        result = GetLastError();
                        if (result != ERROR_IO_PENDING)
                        {
                            returnCode = result;
                            goto ConnectionCleanup;
                        }
                    }
                    else
                        SetEvent(pConnContext->events[EVENT_WRITE_BRIDGE]);
                }
                else if (IRConnected)
                {
                    closesocket(connSock);
                    connSock = INVALID_SOCKET;

                    IRConnected = FALSE;

                    // Benachrichtigung an Bridge-Device senden
                    status = IRCOMM_STATUS_DISCONNECTED;
                    DeviceIoControl(pConnContext->bridgeDev, IOCTL_IRCOMM_REPORT_STATUS,
                                    &status, sizeof(status), NULL, 0, &IOCTLSize, NULL);
                }

                break;*/

            // Lesen von Bridge
            case WAIT_OBJECT_0 + EVENT_READ_BRIDGE:
                ResetEvent(pConnContext->events[EVENT_READ_BRIDGE]);

                if (GetOverlappedResult(pConnContext->bridgeDev, &ovReadBridge,
                                        &transferBrdg2IR, FALSE) && IRConnected)
                {
                    IrCOMMWrite(bufferBrdg2IR, transferBrdg2IR);
                }

                if (IRConnected)
                {
                    ovReadBridge.Offset     = 0;
                    ovReadBridge.OffsetHigh = 0;
                    if (!ReadFile(pConnContext->bridgeDev, bufferBrdg2IR, BRIDGE_BUFFER_SIZE,
                                  &transferBrdg2IR, &ovReadBridge))
                    {
                        result = GetLastError();
                        if (result != ERROR_IO_PENDING)
                        {
                            returnCode = result;
                            goto ConnectionCleanup;
                        }
                    }
                    else
                        SetEvent(pConnContext->events[EVENT_READ_BRIDGE]);
                }

                break;

/*            // Schreiben auf Bridge
            case WAIT_OBJECT_0 + EVENT_WRITE_BRIDGE:
                ResetEvent(pConnContext->events[EVENT_WRITE_BRIDGE]);

                ovReadIR.Offset     = 0;
                ovReadIR.OffsetHigh = 0;
                if (!ReadFile((HANDLE)connSock, bufferIR2Brdg, BRIDGE_BUFFER_SIZE,
                              &transferIR2Brdg, &ovReadIR))
                {
                    result = GetLastError();
                    if ((result != ERROR_IO_PENDING) && IRConnected)
                    {
                        closesocket(connSock);
                        connSock = INVALID_SOCKET;

                        IRConnected = FALSE;

                        // Benachrichtigung an Bridge-Device senden
                        status = IRCOMM_STATUS_DISCONNECTED;
                        DeviceIoControl(pConnContext->bridgeDev, IOCTL_IRCOMM_REPORT_STATUS,
                                        &status, sizeof(status), NULL, 0, &IOCTLSize, NULL);
                    }
                }
                else
                    SetEvent(pConnContext->events[EVENT_READ_IR]);

                break;*/

            default:
                returnCode = GetLastError();
                goto ConnectionCleanup;
        }
    }

  ConnectionCleanup:
    CancelIo(pConnContext->bridgeDev);

    if (IRConnected)
    {
        status = IRCOMM_STATUS_DISCONNECTED;
        DeviceIoControl(pConnContext->bridgeDev, IOCTL_IRCOMM_REPORT_STATUS, &status,
                        sizeof(status), NULL, 0, &IOCTLSize, NULL);
    }

    if (bufferIR2Brdg != NULL)
        free(bufferIR2Brdg);

    return returnCode;
}

int ircomm_tty_data_indication(void *instance, void *sap, struct sk_buff *skb)
{
    OVERLAPPED          ovWriteBridge;
/*    __u8* buf;
    int len;
    len = skb->len;
    buf = skb->data;
    printf("IrCOMM Data: ");
    while (len-- > 0)
        printf("%02X ", *buf++);
    printf("\n");
    len = skb->len;
    buf = skb->data;
    printf("             ");
    while (len-- > 0)
        printf(" %c ", *buf++);
    printf("\n");*/
    ovWriteBridge.Offset     = 0;
    ovWriteBridge.OffsetHigh = 0;
    ovWriteBridge.hEvent     = 0;
    WriteFile(connContext.bridgeDev, skb->data, skb->len, NULL, &ovWriteBridge);

    dev_kfree_skb(skb);

    return 0;
}

void ircomm_tty_check_modem_status(struct ircomm_tty_cb *self)
{
//    printf("****** %08X %08X ******\n", self->settings.dte, self->settings.dce);
}

int main(int argc, char* argv[])
{
    COMMTIMEOUTS            timeout;
    OVERLAPPED              readOvr;
    HANDLE                  waitHandles[2];
    INPUT_RECORD            conin_rec;
    char                    c;
    DWORD                   bytes;
    iobuff_t                rx_buff;
    struct ircomm_tty_cb*   tty_cb;


    printf("Linux IrDA Protocol Stack for Windows, (c) 2002 Jan Kiszka\n");

    if (argc != 2)
    {
        printf("\nusage: irdalnx <port>\n\n\t<port>\tserial port the dongle is connected to\n");
        return 1;
    }

    globalLock = CreateMutex(NULL, false, NULL);
    xmitLock   = CreateMutex(NULL, false, NULL);

    irlmp_init();
    strcpy(sysctl_devname, "Linux-IrDA");
    irlap_init();
    iriap_init();
    irttp_init();
    ircomm_init();
    ircomm_tty = hashbin_new(HB_LOCAL);
    tty_cb = malloc(sizeof(struct ircomm_tty_cb));
    memset(tty_cb, 0, sizeof(struct ircomm_tty_cb));
    init_timer(&tty_cb->watchdog_timer);
    hashbin_insert(ircomm_tty, (irda_queue_t*)tty_cb, /*line*/0, NULL);

    hComPort = CreateFile(argv[1], GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, NULL);
    if (hComPort == INVALID_HANDLE_VALUE)
    {
        printf("\nerror opening serial port %s\n", argv[1]);
        return 1;
    }

    rx_buff.truesize = 4000;
    rx_buff.head     = (__u8*)malloc(4000);
    rx_buff.in_frame = FALSE; 
    rx_buff.state    = OUTSIDE_FRAME; 
    rx_buff.data     = rx_buff.head; 

    timeout.ReadIntervalTimeout = 0;
    timeout.ReadTotalTimeoutConstant = MAXDWORD;
    timeout.ReadTotalTimeoutMultiplier = 0;
    timeout.WriteTotalTimeoutConstant = MAXDWORD;
    timeout.WriteTotalTimeoutMultiplier = 0;
    SetCommTimeouts(hComPort, &timeout);
    GetCommState(hComPort, &dcb);
    dcb.ByteSize     = 8;
    dcb.fParity      = false;
    dcb.fOutxCtsFlow = false;
    dcb.fOutxDsrFlow = false;
    dcb.fDtrControl  = DTR_CONTROL_ENABLE;
    dcb.fRtsControl  = RTS_CONTROL_ENABLE;
    dcb.StopBits     = ONESTOPBIT;

    readOvr.hEvent      = CreateEvent(NULL, true, false, NULL);
    readOvr.Offset      = 0;
    readOvr.OffsetHigh  = 0;

    writeOvr.hEvent     = CreateEvent(NULL, true, false, NULL);
    writeOvr.Offset     = 0;
    writeOvr.OffsetHigh = 0;

    ReadFile(hComPort, &c, 1, NULL, &readOvr);

    waitHandles[0] = readOvr.hEvent;
    waitHandles[1] = GetStdHandle(STD_INPUT_HANDLE);

    sysctl_discovery = 0;

    irda_init_max_qos_capabilies(&qos);
    qos.baud_rate.bits = IR_9600|IR_19200|IR_38400|IR_57600|IR_115200;
    qos.min_turn_time.bits = 0x7f;
    irda_qos_bits_to_value(&qos);
    pIrlap = irlap_open(&netdev, &qos, "ACT-IR220L+");

    CreateThread(NULL, 0, ConnectionThread, &connContext, 0, NULL);

    while (!stop)
    {
        switch (WaitForMultipleObjects(2, waitHandles, false, INFINITE))
        {
            case WAIT_OBJECT_0:
                if (!GetOverlappedResult(hComPort, &readOvr, &bytes, false))
                    printf("^read error %d\n", GetLastError());
                else
                    async_unwrap_char(NULL, NULL, &rx_buff, c);
                ReadFile(hComPort, &c, 1, NULL, &readOvr);
                break;

            case WAIT_OBJECT_0+1:
                if ((ReadConsoleInput(waitHandles[1], &conin_rec, 1, &bytes)) &&
                    (conin_rec.EventType == KEY_EVENT) && conin_rec.Event.KeyEvent.bKeyDown)
                {
                    switch (conin_rec.Event.KeyEvent.uChar.AsciiChar)
                    {
                        case 27:
                            stop = true;
                            break;
                        case 'o':
                            OpenPort();
                            break;
                        case 'c':
                            ClosePort();
                            break;
                        case 'w':
                            IrCOMMWrite("ati9\r\n", 6);
                            break;
                    }
                }
                break;
        }
    }

//    CleanupTimers();

    irttp_cleanup();
    iriap_cleanup();
    irlap_cleanup();
    irlmp_cleanup();
//    hashbin_delete(ircomm_tty, (FREE_FUNC)kfree);

    CloseHandle(globalLock);
    CloseHandle(xmitLock);
    CloseHandle(writeOvr.hEvent);

    CloseHandle(hComPort);
    CloseHandle(waitHandles[0]);
    CloseHandle(waitHandles[1]);

    return 0;
}
