#include <asm/byteorder.h>

#include <linux/netdevice.h>
#include <linux/sched.h>

#include <net/irda/irda.h>
#include <net/irda/irlap.h>
#include <net/irda/timer.h>

#include <net/irda/ircomm_tty.h>
#include <net/irda/ircomm_tty_attach.h>

int tq_immediate;

void irda_notify_init(notify_t *notify)
{
    notify->data_indication = NULL;
    notify->udata_indication = NULL;
    notify->connect_confirm = NULL;
    notify->connect_indication = NULL;
    notify->disconnect_indication = NULL;
    notify->flow_indication = NULL;
    notify->status_indication = NULL;
    notify->instance = NULL;
    strncpy(notify->name, "Unknown", NOTIFY_MAX_NAME);
}

void irda_device_set_media_busy(struct net_device *dev, int status) 
{
	struct irlap_cb *self;

	IRDA_DEBUG(4, __FUNCTION__ "(%s)\n", status ? "TRUE" : "FALSE");

	self = (struct irlap_cb *) dev->atalk_ptr;

	ASSERT(self != NULL, return;);
	ASSERT(self->magic == LAP_MAGIC, return;);

	if (status) {
		self->media_busy = TRUE;
		if (status == SMALL)
			irlap_start_mbusy_timer(self, SMALLBUSY_TIMEOUT);
		else
			irlap_start_mbusy_timer(self, MEDIABUSY_TIMEOUT);
		IRDA_DEBUG( 4, "Media busy!\n");
	} else {
		self->media_busy = FALSE;
		irlap_stop_mbusy_timer(self);
	}
}

int irda_device_is_receiving(struct net_device *dev)
{
	struct if_irda_req req;
	int ret;

	IRDA_DEBUG(2, __FUNCTION__ "()\n");

	if (!dev->do_ioctl) {
		ERROR(__FUNCTION__ "(), do_ioctl not impl. by "
		      "device driver\n");
		return -1;
	}

#define SIOCDEVPRIVATE 0
	ret = dev->do_ioctl(dev, (struct ifreq *) &req, SIOCGRECEIVING);
	if (ret < 0)
		return ret;

	return req.ifr_receiving;
}

extern void ircomm_tty_check_modem_status(struct ircomm_tty_cb *self);

extern int ircomm_tty_data_indication(void *instance, void *sap, struct sk_buff *skb);
/*static int ircomm_tty_data_indication(void *instance, void *sap, struct sk_buff *skb)
{
    __u8* buf;
    int len;
    len = skb->len;
    buf = skb->data;
    printf("IrCOMM Data: ");
    while (len-- > 0)
        printf("%02X ", *buf++);
    printf("\n");
    len = skb->len;
    buf = skb->data;
    printf("             ");
    while (len-- > 0)
        printf(" %c ", *buf++);
    printf("\n");

    dev_kfree_skb(skb);

    return 0;
}*/

int ircomm_tty_control_indication(void *instance, void *sap, struct sk_buff *skb)
{
    struct ircomm_tty_cb *self = (struct ircomm_tty_cb *) instance;
    int clen;

    IRDA_DEBUG(4, __FUNCTION__"()\n");

    ASSERT(self != NULL, return -1;);
    ASSERT(self->magic == IRCOMM_TTY_MAGIC, return -1;);
    ASSERT(skb != NULL, return -1;);

    clen = skb->data[0];

    irda_param_extract_all(self, skb->data+1, IRDA_MIN(skb->len-1, clen),
                           &ircomm_param_info);
    dev_kfree_skb(skb);

    return 0;
}

void ircomm_tty_flow_indication(void *instance, void *sap, LOCAL_FLOW cmd)
{
}

int ircomm_tty_startup(struct ircomm_tty_cb *self)
{
	notify_t notify;
	int ret;

	IRDA_DEBUG(2, __FUNCTION__ "()\n");

	ASSERT(self != NULL, return -1;);
	ASSERT(self->magic == IRCOMM_TTY_MAGIC, return -1;);

	/* Already open */
	if (self->flags & ASYNC_INITIALIZED) {
		IRDA_DEBUG(2, __FUNCTION__ "(), already open so break out!\n");
		return 0;
	}

	/* Register with IrCOMM */
	irda_notify_init(&notify);
	/* These callbacks we must handle ourselves */
	notify.data_indication       = ircomm_tty_data_indication;
	notify.udata_indication      = ircomm_tty_control_indication;
 	notify.flow_indication       = ircomm_tty_flow_indication;

	/* Use the ircomm_tty interface for these ones */
 	notify.disconnect_indication = ircomm_tty_disconnect_indication;
	notify.connect_confirm       = ircomm_tty_connect_confirm;
 	notify.connect_indication    = ircomm_tty_connect_indication;
	strncpy(notify.name, "ircomm_tty", NOTIFY_MAX_NAME);
	notify.instance = self;

	if (!self->ircomm) {
		self->ircomm = ircomm_open(&notify, self->service_type, 
					   self->line);
	}
	if (!self->ircomm)
		return -ENODEV;

	self->slsap_sel = self->ircomm->slsap_sel;

	/* Connect IrCOMM link with remote device */
	ret = ircomm_tty_attach_cable(self);
	if (ret < 0) {
		ERROR(__FUNCTION__ "(), error attaching cable!\n");
		return ret;
	}

	self->flags |= ASYNC_INITIALIZED;

	return 0;
}

void ircomm_tty_shutdown(struct ircomm_tty_cb *self)
{
    unsigned long flags;

    ASSERT(self != NULL, return;);
    ASSERT(self->magic == IRCOMM_TTY_MAGIC, return;);

    IRDA_DEBUG(0, __FUNCTION__ "()\n");

    if (!(self->flags & ASYNC_INITIALIZED))
        return;

    save_flags(flags);
    cli();

    del_timer(&self->watchdog_timer);

    /* Free parameter buffer */
    if (self->ctrl_skb) {
        dev_kfree_skb(self->ctrl_skb);
        self->ctrl_skb = NULL;
    }

    /* Free transmit buffer */
    if (self->tx_skb) {
    	dev_kfree_skb(self->tx_skb);
    	self->tx_skb = NULL;
    }

    ircomm_tty_detach_cable(self);

    if (self->ircomm) {
        ircomm_close(self->ircomm);
        self->ircomm = NULL;
    }
    self->flags &= ~ASYNC_INITIALIZED;

    restore_flags(flags);
}